/*******************************************************************************
*
* FILE:         aesapi.c
*
* DESCRIPTION:  AES-API calls for Crypton Version 0.5
*
* AUTHOR:       Chae Hoon Lim and Hyo Sun Hwang
*               E-mail: chlim@future.co.kr, hyosun@future.co.kr
*               http://crypt.future.co.kr/~chlim
*               Information and Communications Research Center,
*               Future Systems, Inc.
*               372-2, Yang Jae-Dong, Seo Cho-Gu, Seoul, 137-130, KOREA
*               Tel: +82-2-578-0581 (ext. 557), Fax: +82-2-578-0584
*
* DATE:         July 10, 1998
*
* NOTE: This code contains 4 AES-API functions (makeKey, cipherInit,
*       blockEncrypt and blockDecrypt). These functions are using the 4 CRYPTON
*       core functions contained in crypton.c (CryptonEncrypt, CryptonExpandKey, 
*       CryptonGenEncRoundKey and CryptonGenDecRoundkey) as subroutines.
*
*******************************************************************************/

#include "aesapi.h"


/***********************************************************************/
int makeKey(key, direction, keyLen, keyMaterial)
keyInstance *key;
BYTE direction;
int keyLen;     /* in bits */
char *keyMaterial;
{
    int keySize = keyLen/8;
    DWORD ExpandedKey[NoExpandedKeys];

    /* Check if 'direction is invalid (e.g., unknown value)' */
    if( (direction!=DIR_ENCRYPT) && (direction!=DIR_DECRYPT) )
        return BAD_KEY_DIR;

    /* Check if 'keyMaterial is invalid (e.g., wrong length)'
       Proper keyLen = 64+8k, k=0,...,24 */
    if( (keyMaterial==NULL) || (keySize<MIN_KEY_SIZE) || 
        (keySize>MAX_KEY_SIZE) || (keyLen%8 != 0))
        return BAD_KEY_MAT;

    /* Set direction and keyMaterial */
    key->direction = direction;
    key->keyLen = keyLen;
    memcpy(key->keyMaterial, keyMaterial, keySize);

    /* Generate ENC/DECrypt Round Keys */
    CryptonExpandKey(ExpandedKey, key->keyMaterial, keySize);

    if( direction==DIR_ENCRYPT )
        CryptonGenEncRoundKey(key->KS, ExpandedKey);
    else
        CryptonGenDecRoundKey(key->KS, ExpandedKey);

    /* O.K. */
    return TRUE;
}

/***********************************************************************/
int cipherInit(cipher, mode, IV)
cipherInstance *cipher;
BYTE mode;
char *IV;
{
    /* Check if 'the mode passed is unknown' */
    if( (mode!=MODE_ECB) && (mode!=MODE_CBC) && (mode!=MODE_CFB1) )
        return BAD_CIPHER_MODE;

    /* Set mode and IV (IV : in proper case only) */
    cipher->mode = mode;
    if( mode==MODE_CBC || mode==MODE_CFB1 ) {
        BlockCopy(cipher->IV, IV);
    }

    /* O.K. */
    return TRUE;
}

/***********************************************************************/
int blockEncrypt(cipher, key, input, inputLen, outBuffer)
cipherInstance *cipher;
keyInstance *key;
BYTE *input;
int inputLen;
BYTE *outBuffer;
{
    if(key->direction!=DIR_ENCRYPT)
        return BAD_KEY_MAT;

    switch( cipher->mode ) {
        case MODE_ECB:
            {
                int outputLen=inputLen;

                if( (inputLen%BLOCK_LEN)!=0 )
                    return BAD_INPUT_LEN;

                while( inputLen!=0 ) {
                    BlockCopy(outBuffer, input);
                    CryptonEncrypt(outBuffer, key->KS);
                    outBuffer += BLOCK_SIZE;    /* in bytes */
                    input += BLOCK_SIZE;        /* in bytes */
                    inputLen -= BLOCK_LEN;      /* in bits */
                }

                return outputLen;
            }
        case MODE_CBC:
            {
                int outputLen=inputLen;

                if( (inputLen%BLOCK_LEN)!=0 )
                    return BAD_INPUT_LEN;

                while( inputLen!=0 ) {
                    BlockXor(outBuffer, input, cipher->IV);
                    CryptonEncrypt(outBuffer, key->KS);
                    BlockCopy(cipher->IV, outBuffer);
                    outBuffer += BLOCK_SIZE;    /* in bytes */
                    input += BLOCK_SIZE;        /* in bytes */
                    inputLen -= BLOCK_LEN;      /* in bits */
                }

                return outputLen;
            }
        case MODE_CFB1:
            {
                int i, j;
                BYTE tmpBlock[BLOCK_SIZE], tmpBit;

                for( i=0; i<inputLen; i++) {
                    BlockCopy(tmpBlock, cipher->IV);
                    CryptonEncrypt(tmpBlock, key->KS);
                    tmpBit = (tmpBlock[BLOCK_SIZE-1] >> 7)
                            ^ ( (input[i>>3] >> (i&0x07)) & 0x01 );
                    outBuffer[i>>3] |= tmpBit << (i&0x07);

                    for( j=BLOCK_SIZE-1; j>0; j--) {
                        cipher->IV[j] <<= 1;
                        cipher->IV[j] ^= cipher->IV[j-1]>>7;
                    }
                    cipher->IV[0] <<= 1;
                    cipher->IV[0] ^= tmpBit;
                }

                return inputLen;
            }
        default :
            return BAD_CIPHER_STATE;
    }
}

/***********************************************************************/
int blockDecrypt(cipher, key, input, inputLen, outBuffer)
cipherInstance *cipher;
keyInstance *key;
BYTE *input;
int inputLen;
BYTE *outBuffer;
{
    if(key->direction!=DIR_DECRYPT)
        return BAD_KEY_MAT;

    switch( cipher->mode ) {
        case MODE_ECB:
            {
                int outputLen=inputLen;

                if( (inputLen%BLOCK_LEN)!=0 )
                    return BAD_INPUT_LEN;

                while( inputLen!=0 ) {
                    BlockCopy(outBuffer, input);
                    CryptonEncrypt(outBuffer, key->KS);
                    outBuffer += BLOCK_SIZE;    /* in bytes */
                    input += BLOCK_SIZE;        /* in bytes */
                    inputLen -= BLOCK_LEN;      /* in bits */
                }

                return outputLen;
            }
        case MODE_CBC:
            {
                int outputLen=inputLen;

                if( (inputLen%BLOCK_LEN)!=0 )
                    return BAD_INPUT_LEN;

                while( inputLen!=0 ) {
                    BlockCopy(outBuffer, input);
                    CryptonEncrypt(outBuffer, key->KS);
                    BlockXor(outBuffer, outBuffer, cipher->IV);
                    BlockCopy(cipher->IV, input);
                    outBuffer += BLOCK_SIZE;    /* in bytes */
                    input += BLOCK_SIZE;        /* in bytes */
                    inputLen -= BLOCK_LEN;      /* in bits */
                }

                return outputLen;
            }
        case MODE_CFB1:
            {
                int i, j;
                BYTE tmpBlock[BLOCK_SIZE], tmpBit;

                for( i=0; i<inputLen; i++) {
                    BlockCopy(tmpBlock, cipher->IV);
                    CryptonEncrypt(tmpBlock, key->KS);
                    tmpBit = (tmpBlock[BLOCK_SIZE-1] >> 7)
                            ^ ( (input[i>>3] >> (i&0x07)) & 0x01 );
                    outBuffer[i>>3] |= tmpBit << (i&0x07);
                    tmpBit = tmpBlock[BLOCK_SIZE-1] >> 7;

                    for( j=BLOCK_SIZE-1; j>0; j--) {
                        cipher->IV[j] <<= 1;
                        cipher->IV[j] ^= cipher->IV[j-1]>>7;
                    }
                    cipher->IV[0] <<= 1;
                    cipher->IV[0] ^= tmpBit;
                }

                return inputLen;
            }
        default :
            return BAD_CIPHER_STATE;
    }
}

/***************************** END of AESAPI.c *******************************/
