/*******************************************************************************
*
* FILE:         cryptonref.c
*
* DESCRIPTION:  Reference C implementation of Crypton Version 0.7
*
* AUTHOR:       Chae Hoon Lim and Hyo Sun Hwang
*               E-mail: chlim@future.co.kr, hyosun@future.co.kr
*               http://crypt.future.co.kr/~chlim
*               Information and Communications Research Center, 
*               Future Systems, Inc.
*               372-2, Yang Jae-Dong, Seo Cho-Gu, Seoul, 137-130, KOREA
*               Tel: +82-2-578-0581 (ext. 557), Fax: +82-2-578-0584
*
* DATE:         July 10, 1998
*
* Notes: This code is a reference implementation written to help the reader 
*        to understand CRYPTON. No optimization tried. 
*        It does exactly the same computation as "crypton.c" (but much slower).
*
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "crypton.h"
#include "crypton.tab"  /* S-boxes */


/*############# DEFINITIONS FOR ENCRYPTION/DECRYPTION ############*/

/* byte extracting/positioning from/into DWORD */
#ifdef BIG_ENDIAN
    #define GetByte0(A)  ( (BYTE)((A)>>24) )
    #define GetByte1(A)  ( (BYTE)((A)>>16) )
    #define GetByte2(A)  ( (BYTE)((A)>> 8) )
    #define GetByte3(A)  ( (BYTE)((A)    ) )

    #define PutByte0(A)  ( (DWORD)(A)<<24 )
    #define PutByte1(A)  ( (DWORD)(A)<<16 )
    #define PutByte2(A)  ( (DWORD)(A)<< 8 )
    #define PutByte3(A)  ( (DWORD)(A)     )
#else
    #define GetByte0(A)  ( (BYTE)((A)    ) )
    #define GetByte1(A)  ( (BYTE)((A)>> 8) )
    #define GetByte2(A)  ( (BYTE)((A)>>16) )
    #define GetByte3(A)  ( (BYTE)((A)>>24) )

    #define PutByte0(A)  ( (DWORD)(A)     )
    #define PutByte1(A)  ( (DWORD)(A)<< 8 )
    #define PutByte2(A)  ( (DWORD)(A)<<16 )
    #define PutByte3(A)  ( (DWORD)(A)<<24 )
#endif


/* Sbox transformation for ODD rounds */
#define SboxTrans0(B, A) {                    \
    B[0]  = PutByte0(S0[GetByte0(A[0])]) ^    \
            PutByte1(S1[GetByte1(A[0])]) ^    \
            PutByte2(S2[GetByte2(A[0])]) ^    \
            PutByte3(S3[GetByte3(A[0])]);     \
    B[1]  = PutByte0(S1[GetByte0(A[1])]) ^    \
            PutByte1(S2[GetByte1(A[1])]) ^    \
            PutByte2(S3[GetByte2(A[1])]) ^    \
            PutByte3(S0[GetByte3(A[1])]);     \
    B[2]  = PutByte0(S2[GetByte0(A[2])]) ^    \
            PutByte1(S3[GetByte1(A[2])]) ^    \
            PutByte2(S0[GetByte2(A[2])]) ^    \
            PutByte3(S1[GetByte3(A[2])]);     \
    B[3]  = PutByte0(S3[GetByte0(A[3])]) ^    \
            PutByte1(S0[GetByte1(A[3])]) ^    \
            PutByte2(S1[GetByte2(A[3])]) ^    \
            PutByte3(S2[GetByte3(A[3])]);     \
}

/* Sbox transformation for EVEN rounds */
#define SboxTrans1(B, A) {                    \
    B[0]  = PutByte0(S2[GetByte0(A[0])]) ^    \
            PutByte1(S3[GetByte1(A[0])]) ^    \
            PutByte2(S0[GetByte2(A[0])]) ^    \
            PutByte3(S1[GetByte3(A[0])]);     \
    B[1]  = PutByte0(S3[GetByte0(A[1])]) ^    \
            PutByte1(S0[GetByte1(A[1])]) ^    \
            PutByte2(S1[GetByte2(A[1])]) ^    \
            PutByte3(S2[GetByte3(A[1])]);     \
    B[2]  = PutByte0(S0[GetByte0(A[2])]) ^    \
            PutByte1(S1[GetByte1(A[2])]) ^    \
            PutByte2(S2[GetByte2(A[2])]) ^    \
            PutByte3(S3[GetByte3(A[2])]);     \
    B[3]  = PutByte0(S1[GetByte0(A[3])]) ^    \
            PutByte1(S2[GetByte1(A[3])]) ^    \
            PutByte2(S3[GetByte2(A[3])]) ^    \
            PutByte3(S0[GetByte3(A[3])]);     \
}


/* Mask values used for bit permutation */
#ifdef BIG_ENDIAN     /* byte-reversed versions */
    #define MW0   0xfcf3cf3f
    #define MW1   0xf3cf3ffc
    #define MW2   0xcf3ffcf3
    #define MW3   0x3ffcf3cf
#else
    #define MW0   0x3fcff3fc
    #define MW1   0xfc3fcff3
    #define MW2   0xf3fc3fcf
    #define MW3   0xcff3fc3f
#endif


/* Bit permutation for ODD rounds */
#define BitPerm0(B, A) {                    \
    B[0] = (A[0] & MW0) ^ (A[1] & MW1) ^    \
           (A[2] & MW2) ^ (A[3] & MW3);     \
    B[1] = (A[0] & MW1) ^ (A[1] & MW2) ^    \
           (A[2] & MW3) ^ (A[3] & MW0);     \
    B[2] = (A[0] & MW2) ^ (A[1] & MW3) ^    \
           (A[2] & MW0) ^ (A[3] & MW1);     \
    B[3] = (A[0] & MW3) ^ (A[1] & MW0) ^    \
           (A[2] & MW1) ^ (A[3] & MW2);     \
}

/* Bit permutation for EVEN rounds */
#define BitPerm1(B, A) {                    \
    B[0] = (A[0] & MW2) ^ (A[1] & MW3) ^    \
           (A[2] & MW0) ^ (A[3] & MW1);     \
    B[1] = (A[0] & MW3) ^ (A[1] & MW0) ^    \
           (A[2] & MW1) ^ (A[3] & MW2);     \
    B[2] = (A[0] & MW0) ^ (A[1] & MW1) ^    \
           (A[2] & MW2) ^ (A[3] & MW3);     \
    B[3] = (A[0] & MW1) ^ (A[1] & MW2) ^    \
           (A[2] & MW3) ^ (A[3] & MW0);     \
}


/* Byte transposition */
#define ByteTrans(B, A) {                                          \
    B[0] = PutByte0(GetByte0(A[0])) ^ PutByte1(GetByte0(A[1])) ^   \
           PutByte2(GetByte0(A[2])) ^ PutByte3(GetByte0(A[3]));    \
    B[1] = PutByte0(GetByte1(A[0])) ^ PutByte1(GetByte1(A[1])) ^   \
           PutByte2(GetByte1(A[2])) ^ PutByte3(GetByte1(A[3]));    \
    B[2] = PutByte0(GetByte2(A[0])) ^ PutByte1(GetByte2(A[1])) ^   \
           PutByte2(GetByte2(A[2])) ^ PutByte3(GetByte2(A[3]));    \
    B[3] = PutByte0(GetByte3(A[0])) ^ PutByte1(GetByte3(A[1])) ^   \
           PutByte2(GetByte3(A[2])) ^ PutByte3(GetByte3(A[3]));    \
}


/* Key addition */
#define KeyAdd(A, K) {      \
    A[0] ^= (K)[0];         \
    A[1] ^= (K)[1];         \
    A[2] ^= (K)[2];         \
    A[3] ^= (K)[3];         \
}


/* Round transformation for odd rounds */
#define CryptonRound0(B, A, K) {    \
    SboxTrans0(B, A);               \
    BitPerm0(A, B);                 \
    ByteTrans(B, A);                \
    KeyAdd(B, K);                   \
}

/* Round transformation for even rounds */
#define CryptonRound1(B, A, K) {    \
    SboxTrans1(B, A);               \
    BitPerm1(A, B);                 \
    ByteTrans(B, A);                \
    KeyAdd(B, K);                   \
}


/* Special Final round: SboxTrans1, ByteTrans and KeyAdd */
#define CryptonFinalR(pb, A, K) {                      \
    pb[ 0] = S2[GetByte0(A[0])] ^ GetByte0((K)[0]);    \
    pb[ 1] = S3[GetByte0(A[1])] ^ GetByte1((K)[0]);    \
    pb[ 2] = S0[GetByte0(A[2])] ^ GetByte2((K)[0]);    \
    pb[ 3] = S1[GetByte0(A[3])] ^ GetByte3((K)[0]);    \
    pb[ 4] = S3[GetByte1(A[0])] ^ GetByte0((K)[1]);    \
    pb[ 5] = S0[GetByte1(A[1])] ^ GetByte1((K)[1]);    \
    pb[ 6] = S1[GetByte1(A[2])] ^ GetByte2((K)[1]);    \
    pb[ 7] = S2[GetByte1(A[3])] ^ GetByte3((K)[1]);    \
    pb[ 8] = S0[GetByte2(A[0])] ^ GetByte0((K)[2]);    \
    pb[ 9] = S1[GetByte2(A[1])] ^ GetByte1((K)[2]);    \
    pb[10] = S2[GetByte2(A[2])] ^ GetByte2((K)[2]);    \
    pb[11] = S3[GetByte2(A[3])] ^ GetByte3((K)[2]);    \
    pb[12] = S1[GetByte3(A[0])] ^ GetByte0((K)[3]);    \
    pb[13] = S2[GetByte3(A[1])] ^ GetByte1((K)[3]);    \
    pb[14] = S3[GetByte3(A[2])] ^ GetByte2((K)[3]);    \
    pb[15] = S0[GetByte3(A[3])] ^ GetByte3((K)[3]);    \
}


/* save inter-round values */
#if defined(CheckInterValue)
   extern BYTE pdwInterVal[NoRounds+1][CryptonBlockSize];
   #define INTER_VALUE_OUT(pdw, N)  { \
       if(PRN)
           memcpy(pdwInterVal[N], (BYTE *)(pdw), CryptonBlockSize); \
   }
#else
   #define INTER_VALUE_OUT(pdw, N)
#endif

/*############### BLOCK ENCRYPTION ROUTINE ###############*/

/* encrypt/decrypt one 128-bit block of data */
void CryptonEncrypt(pbData, pdwRoundKey)
BYTE *pbData;
DWORD *pdwRoundKey;
{
    DWORD BlockP[4], BlockC[4], *K=pdwRoundKey;

    INTER_VALUE_OUT(pbData, 0);
    BlockP[0] = ((DWORD *)pbData)[0] ^ K[0];
    BlockP[1] = ((DWORD *)pbData)[1] ^ K[1];
    BlockP[2] = ((DWORD *)pbData)[2] ^ K[2];
    BlockP[3] = ((DWORD *)pbData)[3] ^ K[3];

    CryptonRound0(BlockC, BlockP, K+ 4);       /*  1 */
    INTER_VALUE_OUT(BlockC, 1);
    CryptonRound1(BlockP, BlockC, K+ 8);       /*  2 */
    INTER_VALUE_OUT(BlockP, 2);
    CryptonRound0(BlockC, BlockP, K+12);       /*  3 */
    INTER_VALUE_OUT(BlockC, 3);
    CryptonRound1(BlockP, BlockC, K+16);       /*  4 */
    INTER_VALUE_OUT(BlockP, 4);
    CryptonRound0(BlockC, BlockP, K+20);       /*  5 */
    INTER_VALUE_OUT(BlockC, 5);
    CryptonRound1(BlockP, BlockC, K+24);       /*  6 */
    INTER_VALUE_OUT(BlockP, 6);
    CryptonRound0(BlockC, BlockP, K+28);       /*  7 */
    INTER_VALUE_OUT(BlockC, 7);
    CryptonRound1(BlockP, BlockC, K+32);       /*  8 */
    INTER_VALUE_OUT(BlockP, 8);
    CryptonRound0(BlockC, BlockP, K+36);       /*  9 */
    INTER_VALUE_OUT(BlockC, 9);
    CryptonRound1(BlockP, BlockC, K+40);       /* 10 */
    INTER_VALUE_OUT(BlockP, 10);
    CryptonRound0(BlockC, BlockP, K+44);       /* 11 */
    INTER_VALUE_OUT(BlockC, 11);
    CryptonFinalR(pbData, BlockC, K+48);       /* 12 */
    INTER_VALUE_OUT(pbData, 12);
}


/*##################### Key Scheduling Routines ##################*/

#define KeyExpand0(B, A, K) {    \
    BitPerm0(B, A);              \
    KeyAdd(B, K);                \
    SboxTrans0(A, B);            \
    ByteTrans(B, A);             \
}

#define KeyExpand1(B, A, K) {    \
    BitPerm1(B, A);              \
    KeyAdd(B, K);                \
    SboxTrans1(A, B);            \
    ByteTrans(B, A);             \
}


static BYTE pbConst[32] = {0x85, 0xae, 0x67, 0xbb, 0x72, 0xf3, 0x6e, 0x3c, 
    0x3a, 0xf5, 0x4f, 0xa5, 0x7f, 0x52, 0x0e, 0x51, 0x8c, 0x68, 0x05, 0x9b, 
    0xab, 0xd9, 0x83, 0x1f, 0x19, 0xcd, 0xe0, 0x5b, 0x5d, 0x9d, 0xbb, 0xcb};

/******************* GENERATE EXPANDED KEYS *********************/ 
/* generate 8 expanded keys from a user-supplied key */
void CryptonExpandKey(pdwExpKey, pbUserKey, dwUserKeyLen)
DWORD *pdwExpKey;
BYTE *pbUserKey;
int dwUserKeyLen;
{
    int i;
    BYTE pbTemp[32];
    DWORD Sum0, Sum1, *EK = pdwExpKey, *pdw;
    DWORD UK0[NoExpandedKeys/2]={0}, UK1[NoExpandedKeys/2]={0};

    BlockSet(pbTemp, 0);
    BlockSet(pbTemp+16, 0);
    memcpy(pbTemp, pbUserKey, dwUserKeyLen);

    pdw = (DWORD *)pbTemp;
    for(i=0; i<dwUserKeyLen/4; ++i) {
        if(i&1) UK1[i/2] = *(pdw+i);  /* odd words */
        else UK0[i/2] = *(pdw+i);     /* even words */
    }

    pdw = (DWORD *)pbConst;
    KeyExpand0( EK   , UK0, pdw  );
    KeyExpand1((EK+4), UK1, pdw+4);
    Sum0 = EK[0] ^ EK[1] ^ EK[2] ^ EK[3];
    Sum1 = EK[4] ^ EK[5] ^ EK[6] ^ EK[7];
    for(i=0; i<4; ++i) {
        EK[i  ] ^= Sum1;
        EK[i+4] ^= Sum0;
    }
}


/****************************************************************/
/* masking words for word permutation */
    #define    M20    0xcffccffc
    #define    M31    0x3ff33ff3
    #define    M02    0xfccffccf
    #define    M13    0xf33ff33f

/* equivalent to ByteTrans, BitPerm and ByteTrans */
#ifdef BIG_ENDIAN
    #define WordPerm0(A0)                               \
        (      (A0 & M02    ) ^ ROTL(A0 & M31,  8) ^   \
          ROTL(A0 & M20, 16) ^ ROTL(A0 & M13, 24) )
    #define WordPerm1(A1)                               \
        (      (A1 & M13    ) ^ ROTL(A1 & M02,  8) ^   \
          ROTL(A1 & M31, 16) ^ ROTL(A1 & M20, 24) )
    #define WordPerm2(A2)                               \
        (      (A2 & M20    ) ^ ROTL(A2 & M13,  8) ^   \
          ROTL(A2 & M02, 16) ^ ROTL(A2 & M31, 24) )
    #define WordPerm3(A3)                               \
        (      (A3 & M31    ) ^ ROTL(A3 & M20,  8) ^   \
          ROTL(A3 & M13, 16) ^ ROTL(A3 & M02, 24) )
#else
    #define WordPerm0(A0)                               \
        (      (A0 & M20    ) ^ ROTL(A0 & M31,  8) ^   \
          ROTL(A0 & M02, 16) ^ ROTL(A0 & M13, 24) )
    #define WordPerm1(A1)                               \
        (      (A1 & M31    ) ^ ROTL(A1 & M02,  8) ^   \
          ROTL(A1 & M13, 16) ^ ROTL(A1 & M20, 24) )
    #define WordPerm2(A2)                               \
        (      (A2 & M02    ) ^ ROTL(A2 & M13,  8) ^   \
          ROTL(A2 & M20, 16) ^ ROTL(A2 & M31, 24) )
    #define WordPerm3(A3)                               \
        (      (A3 & M13    ) ^ ROTL(A3 & M20,  8) ^   \
          ROTL(A3 & M31, 16) ^ ROTL(A3 & M02, 24) )
#endif


#ifdef BIG_ENDIAN
    #define ROTATE(X, n)  ROTR(X, n)
#else
    #define ROTATE(X, n)  ROTL(X, n)
#endif

/* Round key update */
#define RoundKeyUpdate0(RK, EK, rot0, rot1, RC) {     \
    (RK)[0] = ROTATE((EK)[0], rot0);                  \
    (RK)[1] = (EK)[1] ^ RC;                           \
    (RK)[2] = ROTATE((EK)[2], rot1);                  \
    (RK)[3] = (EK)[3] ^ RC;                           \
}

#define RoundKeyUpdate1(RK, EK, RC, rot0, rot1) {     \
    (RK)[0] = (EK)[0] ^ RC;                           \
    (RK)[1] = ROTATE((EK)[1], rot0);                  \
    (RK)[2] = (EK)[2] ^ RC;                           \
    (RK)[3] = ROTATE((EK)[3], rot1);                  \
}


/* Round constants */
#define  RC0  0x01010101UL
#define  RC1  0x02020202UL
#define  RC2  0x04040404UL
#define  RC3  0x08080808UL
#define  RC4  0x10101010UL
#define  RC5  0x20202020UL


/**************** GENERATE ENCRYPTION ROUND KEYS *****************/
void CryptonGenEncRoundKey(pdwEncKey, pdwExpKey)
DWORD *pdwExpKey, *pdwEncKey;
{
    BlockCopy(pdwEncKey  , pdwExpKey  );                      /* round  0 */
    BlockCopy(pdwEncKey+4, pdwExpKey+4);                      /* round  1 */

    RoundKeyUpdate0(pdwEncKey+ 8, pdwEncKey   ,  8, 16, RC0); /* round  2 */
    RoundKeyUpdate1(pdwEncKey+12, pdwEncKey+ 4, RC0, 16, 24); /* round  3 */

    RoundKeyUpdate1(pdwEncKey+16, pdwEncKey+ 8, RC1, 24,  8); /* round  4 */
    RoundKeyUpdate0(pdwEncKey+20, pdwEncKey+12,  8, 16, RC1); /* round  5 */

    RoundKeyUpdate0(pdwEncKey+24, pdwEncKey+16, 16, 24, RC2); /* round  6 */
    RoundKeyUpdate1(pdwEncKey+28, pdwEncKey+20, RC2, 24,  8); /* round  7 */

    RoundKeyUpdate1(pdwEncKey+32, pdwEncKey+24, RC3,  8, 16); /* round  8 */
    RoundKeyUpdate0(pdwEncKey+36, pdwEncKey+28, 16, 24, RC3); /* round  9 */

    RoundKeyUpdate0(pdwEncKey+40, pdwEncKey+32, 24,  8, RC4); /* round 10 */
    RoundKeyUpdate1(pdwEncKey+44, pdwEncKey+36, RC4,  8, 16); /* round 11 */

    RoundKeyUpdate1(pdwEncKey+48, pdwEncKey+40, RC5, 16, 24); /* round 12 */
    pdwEncKey[48] = WordPerm2(pdwEncKey[48]);
    pdwEncKey[49] = WordPerm3(pdwEncKey[49]);
    pdwEncKey[50] = WordPerm0(pdwEncKey[50]);
    pdwEncKey[51] = WordPerm1(pdwEncKey[51]);
}


/**************** GENERATE ENCRYPTION ROUND KEYS *****************/
void CryptonGenDecRoundKey(pdwDecKey, pdwExpKey)
DWORD *pdwExpKey, *pdwDecKey;
{
    pdwDecKey[0] = WordPerm0(pdwExpKey[0]) ^ RC1 ^ RC3 ^ RC5; /* round  0 */
    pdwDecKey[1] = WordPerm1(pdwExpKey[1]) ^ RC0 ^ RC2 ^ RC4;
    pdwDecKey[2] = WordPerm2(pdwExpKey[2]) ^ RC1 ^ RC3 ^ RC5;
    pdwDecKey[3] = WordPerm3(pdwExpKey[3]) ^ RC0 ^ RC2 ^ RC4;

    pdwDecKey[4] = WordPerm3(pdwExpKey[4]) ^ RC0 ^ RC2 ^ RC4; /* round  1 */
    pdwDecKey[5] = WordPerm3(pdwExpKey[5]) ^ RC1 ^ RC3;
    pdwDecKey[6] = WordPerm3(pdwExpKey[6]) ^ RC0 ^ RC2 ^ RC4;
    pdwDecKey[7] = WordPerm1(pdwExpKey[7]) ^ RC1 ^ RC3;

    RoundKeyUpdate1(pdwDecKey+ 8, pdwDecKey   , RC5, 16, 24); /* round  2 */
    RoundKeyUpdate1(pdwDecKey+12, pdwDecKey+ 4, RC4,  8, 16); /* round  3 */

    RoundKeyUpdate0(pdwDecKey+16, pdwDecKey+ 8, 24,  8, RC4); /* round  4 */
    RoundKeyUpdate0(pdwDecKey+20, pdwDecKey+12, 16, 24, RC3); /* round  5 */

    RoundKeyUpdate1(pdwDecKey+24, pdwDecKey+16, RC3,  8, 16); /* round  6 */
    RoundKeyUpdate1(pdwDecKey+28, pdwDecKey+20, RC2, 24,  8); /* round  7 */

    RoundKeyUpdate0(pdwDecKey+32, pdwDecKey+24, 16, 24, RC2); /* round  8 */
    RoundKeyUpdate0(pdwDecKey+36, pdwDecKey+28,  8, 16, RC1); /* round  9 */

    RoundKeyUpdate1(pdwDecKey+40, pdwDecKey+32, RC1, 24,  8); /* round 10 */
    RoundKeyUpdate1(pdwDecKey+44, pdwDecKey+36, RC0, 16, 24); /* round 11 */

    /* Last round keys: round 12 */
    pdwDecKey[48] = pdwExpKey[0]; 
    pdwDecKey[49] = pdwExpKey[1];
    pdwDecKey[50] = pdwExpKey[2]; 
    pdwDecKey[51] = pdwExpKey[3]; 
}


/************************ Generate CRYPTON S-boxes *********************/
 
void CryptonGenSbox(void)
{
    int i, R, L;
    static BYTE P0[16] = {15, 5,10, 1, 5, 5, 8, 9,10, 2,10, 3, 4, 6,12,15};
    static BYTE P1[16] = {10,15, 4,13, 5, 8,14,12, 3, 9, 6, 2, 7, 1,11, 0};
    static BYTE P2[16] = { 0, 4, 2, 4, 1,15, 2,14, 8, 8,15,13, 1,11, 7,15};

    extern BYTE S0[256], S1[256];
    extern DWORD SS0[256], SS1[256], SS2[256], SS3[256];


    for(i=0; i<256; ++i) {
        R = i&0xf;
        L = (i>>4) ^ P0[R];
        R ^= P1[L];
        L ^= P2[R];
        R ^= (L<< 4);
        L = ((R<<6) ^ (R>>2)) & 0xff;

         S0[i] = (BYTE)R;
        SS0[i] = PutByte3(S0[R]&0x3f) ^ PutByte2(S0[R]&0xcf) ^
                 PutByte1(S0[R]&0xf3) ^ PutByte0(S0[R]&0xfc);
         S1[i] = (BYTE)L;
        SS1[i] = PutByte3(S1[L]&0xfc) ^ PutByte2(S1[L]&0x3f) ^
                 PutByte1(S1[L]&0xcf) ^ PutByte0(S1[L]&0xf3);
         S2[R] = (BYTE)i;
        SS2[R] = PutByte3(S2[i]&0xf3) ^ PutByte2(S2[i]&0xfc) ^
                 PutByte1(S2[i]&0x3f) ^ PutByte0(S2[i]&0xcf);
         S3[L] = (BYTE)i;
        SS3[L] = PutByte3(S3[i]&0xcf) ^ PutByte2(S3[i]&0xf3) ^
                 PutByte1(S3[i]&0xfc) ^ PutByte0(S3[i]&0x3f);
    }
}

/******************** END OF CRYPTONREF.C *********************/
