#include <stdlib.h>
#include <string.h>

#include "rijndael-alg-fst.h"
#include "rijndael-api-fst.h"


#define CHECK_LENGTH(x) ((x) == 128 || (x) == 192 || (x) == 256)


int makeKey(keyInstance *key, BYTE direction, int keyLen, int blockLen, char *keyMaterial)
{
	word8 k[MAXKC][4];
	int i, j;
	
	if (key == NULL) {
		return BAD_KEY_INSTANCE;
	}

	if ((direction == DIR_ENCRYPT) || (direction == DIR_DECRYPT)) {
		key->direction = direction;
	} else {
		return BAD_KEY_DIR;
	}

	if ((keyLen == 128) || (keyLen == 192) || (keyLen == 256)) { 
		key->keyLen = keyLen;
	} else {
		return BAD_KEY_MAT;
	}

	if ( keyMaterial ) {
		strncpy(key->keyMaterial, keyMaterial, keyLen/4);
	}

	key->blockLen = blockLen;

	/* initialize key schedule: */
 	for(i = 0; i < key->keyLen/8; i++) {
		sscanf(key->keyMaterial + 2*i,"%2x", &j);
		k[i / 4][i % 4] = (word8) j; 
	}
	rijndaelKeySched (k, key->keyLen, key->blockLen, key->keySched);
	if (direction == DIR_DECRYPT)
		rijndaelKeyEnctoDec (key->keyLen, key->blockLen, key->keySched);

	return TRUE;
}

int cipherInit(cipherInstance *cipher, BYTE mode, int blockLen, char *IV)
{
	int i, j;
	
	if ((mode == MODE_ECB) || (mode == MODE_CBC) || (mode == MODE_CFB1)) {
		cipher->mode = mode;
	} else {
		return BAD_CIPHER_MODE;
	}
	
	if ((blockLen == 128) || (blockLen == 192) || (blockLen == 256)) { 
		cipher->blockLen = blockLen;
	} else {
		return BAD_BLOCK_LENGTH;
	}

	if (IV != NULL) {
 		for(i = 0; i < blockLen/8; i++) {
			sscanf(IV + 2*i,"%2x",&j);
			cipher->IV[i] = (word8) j;
		} 
	}

	return TRUE;
}


int blockEncrypt(cipherInstance *cipher,
	keyInstance *key, BYTE *input, int inputLen, BYTE *outBuffer)
{
	int i, j, k, numBlocks;
	word8 block[MAXBC][4], iv[MAXBC][4];

	if (cipher == NULL ||
		key == NULL ||
		key->direction == DIR_DECRYPT ||
		cipher->blockLen != key->blockLen) {
		return BAD_CIPHER_STATE;
	}
	

	numBlocks = inputLen/cipher->blockLen;
	
	switch (cipher->mode) {
	case MODE_ECB: 
		for (i = numBlocks; i > 0; i--) {
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse input stream into rectangular array */
				*((word32*)block[j]) = *((word32*)(input+4*j));
			}
			rijndaelEncrypt (block, key->keyLen, cipher->blockLen, key->keySched);
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse rectangular array into output ciphertext bytes */
				*((word32*)(outBuffer+4*j)) = *((word32*)block[j]);
			}
		}
		break;
		
	case MODE_CBC:
  		for (j = cipher->blockLen/32-1; j >= 0; j--) {
			/* parse initial value into rectangular array */
			*((word32*)block[j]) = *((word32*)(cipher->IV+4*j));
		}
		for (i = numBlocks; i > 0; i--) {
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse input stream into rectangular array and exor with 
				   IV or the previous ciphertext */
				*((word32*)block[j]) ^= *((word32*)(input+4*j));
			}
			rijndaelEncrypt (block, key->keyLen, cipher->blockLen, key->keySched);
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse rectangular array into output ciphertext bytes */
				*((word32*)(outBuffer+4*j)) = *((word32*)block[j]);
			}
		}
		break;
	
	case MODE_CFB1:
  		for (j = cipher->blockLen/32-1; j >= 0; j--) {
			/* parse initial value into iv rectangular array */
			*((word32*)iv[j]) = *((word32*)(cipher->IV+4*j));
		}
		for (i = numBlocks; i > 0; i--) {
			for (k = 0; k < cipher->blockLen; k++) {
				for (j = cipher->blockLen/32-1; j >= 0; j--) {
					/* parse iv stream into rectangular array */
					*((word32*)block[j]) = *((word32*)iv[j]);
				}
				rijndaelEncrypt (block, key->keyLen, cipher->blockLen, key->keySched);
				outBuffer[k/8] ^= (block[0][0] & 0x80) >> (k & 7);
				iv[0][0] = (iv[0][0] << 1) | (iv[0][1] >> 7);
				iv[0][1] = (iv[0][1] << 1) | (iv[0][2] >> 7);
				iv[0][2] = (iv[0][2] << 1) | (iv[0][3] >> 7);
				iv[0][3] = (iv[0][3] << 1) | (iv[1][0] >> 7);
				iv[1][0] = (iv[1][0] << 1) | (iv[1][1] >> 7);
				iv[1][1] = (iv[1][1] << 1) | (iv[1][2] >> 7);
				iv[1][2] = (iv[1][2] << 1) | (iv[1][3] >> 7);
				iv[1][3] = (iv[1][3] << 1) | (iv[2][0] >> 7);
				iv[2][0] = (iv[2][0] << 1) | (iv[2][1] >> 7);
				iv[2][1] = (iv[2][1] << 1) | (iv[2][2] >> 7);
				iv[2][2] = (iv[2][2] << 1) | (iv[2][3] >> 7);
				iv[2][3] = (iv[2][3] << 1) | (iv[3][0] >> 7);
				iv[3][0] = (iv[3][0] << 1) | (iv[3][1] >> 7);
				iv[3][1] = (iv[3][1] << 1) | (iv[3][2] >> 7);
				iv[3][2] = (iv[3][2] << 1) | (iv[3][3] >> 7);
				iv[3][3] = (iv[3][3] << 1) | (outBuffer[k/8] >> (7-(k&7))) & 1;
			}
		}
		break;
	
	default:
		return BAD_CIPHER_STATE;
	}
	
	return numBlocks*cipher->blockLen;
}

int blockDecrypt(cipherInstance *cipher,
	keyInstance *key, BYTE *input, int inputLen, BYTE *outBuffer)
{
	int i, j, k, numBlocks;
	word8 block[MAXBC][4], iv[MAXBC][4];

	if (cipher == NULL ||
		key == NULL ||
		cipher->mode != MODE_CFB1 && key->direction == DIR_ENCRYPT ||
		cipher->blockLen != key->blockLen) {
		return BAD_CIPHER_STATE;
	}
	

	numBlocks = inputLen/cipher->blockLen;
	
	switch (cipher->mode) {
	case MODE_ECB: 
		for (i = numBlocks; i > 0; i--) {
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse input stream into rectangular array */
				*((word32*)block[j]) = *((word32*)(input+4*j));
			}
			rijndaelDecrypt (block, key->keyLen, cipher->blockLen, key->keySched);
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse rectangular array into output ciphertext bytes */
				*((word32*)(outBuffer+4*j)) = *((word32*)block[j]);
			}
		}
		break;
		
	case MODE_CBC:
		/* first block */
  		for (j = cipher->blockLen/32-1; j >= 0; j--) {
			/* parse input stream into rectangular array */
  			*((word32*)block[j]) = *((word32*)(input+4*j));
		}
		rijndaelDecrypt (block, key->keyLen, cipher->blockLen, key->keySched);
		
		for (j = cipher->blockLen/32-1; j >= 0; j--) {
			/* exor the IV and parse rectangular array into output ciphertext bytes */
  			*((word32*)(outBuffer+4*j)) = *((word32*)block[j]) ^ *((word32*)(cipher->IV+4*j));
		}
		
		/* next blocks */
		for (i = numBlocks-1; i > 0; i--) {
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* parse input stream into rectangular array */
	  			*((word32*)block[j]) = *((word32*)(input+cipher->blockLen/8+4*j));
			}
			rijndaelDecrypt (block, key->keyLen, cipher->blockLen, key->keySched);
			
			for (j = cipher->blockLen/32-1; j >= 0; j--) {
				/* exor previous ciphertext block and parse rectangular array 
				       into output ciphertext bytes */
				*((word32*)(outBuffer+cipher->blockLen/8+4*j)) = *((word32*)block[j]) ^
					*((word32*)(input-cipher->blockLen/8+4*j));
			}
		}
		break;
	
	case MODE_CFB1:
  		for (j = cipher->blockLen/32-1; j >= 0; j--) {
			/* parse initial value into iv rectangular array */
			*((word32*)iv[j]) = *((word32*)(cipher->IV+4*j));
		}
		for (i = numBlocks; i > 0; i--) {
			for (k = 0; k < cipher->blockLen; k++) {
				for (j = cipher->blockLen/32-1; j >= 0; j--) {
					/* parse iv stream into rectangular array */
					*((word32*)block[j]) = *((word32*)iv[j]);
				}
				rijndaelEncrypt (block, key->keyLen, cipher->blockLen, key->keySched);
				iv[0][0] = (iv[0][0] << 1) | (iv[0][1] >> 7);
				iv[0][1] = (iv[0][1] << 1) | (iv[0][2] >> 7);
				iv[0][2] = (iv[0][2] << 1) | (iv[0][3] >> 7);
				iv[0][3] = (iv[0][3] << 1) | (iv[1][0] >> 7);
				iv[1][0] = (iv[1][0] << 1) | (iv[1][1] >> 7);
				iv[1][1] = (iv[1][1] << 1) | (iv[1][2] >> 7);
				iv[1][2] = (iv[1][2] << 1) | (iv[1][3] >> 7);
				iv[1][3] = (iv[1][3] << 1) | (iv[2][0] >> 7);
				iv[2][0] = (iv[2][0] << 1) | (iv[2][1] >> 7);
				iv[2][1] = (iv[2][1] << 1) | (iv[2][2] >> 7);
				iv[2][2] = (iv[2][2] << 1) | (iv[2][3] >> 7);
				iv[2][3] = (iv[2][3] << 1) | (iv[3][0] >> 7);
				iv[3][0] = (iv[3][0] << 1) | (iv[3][1] >> 7);
				iv[3][1] = (iv[3][1] << 1) | (iv[3][2] >> 7);
				iv[3][2] = (iv[3][2] << 1) | (iv[3][3] >> 7);
				iv[3][3] = (iv[3][3] << 1) | (input[k/8] >> (7-(k&7))) & 1;
				outBuffer[k/8] ^= (block[0][0] & 0x80) >> (k & 7);
			}
		}
		break;

	default:
		return BAD_CIPHER_STATE;
	}
	
	return numBlocks*cipher->blockLen;
}


/**
 *	cipherUpdateRounds:
 *
 *	Encrypts/Decrypts exactly one full block a specified number of rounds.
 *	Only used in the Intermediate Value Known Answer Test.	
 *
 *	Returns:
 *		TRUE - on success
 *		BAD_CIPHER_STATE - cipher in bad state (e.g., not initialized)
 */
int cipherUpdateRounds(cipherInstance *cipher,
	keyInstance *key, BYTE *input, int inputLen, BYTE *outBuffer, int rounds)
{
	int j;
	word8 block[MAXBC][4];

	if (cipher == NULL ||
		key == NULL ||
		cipher->blockLen != key->blockLen) {
		return BAD_CIPHER_STATE;
	}

	for (j = cipher->blockLen/32-1; j >= 0; j--) {
		/* parse input stream into rectangular array */
  		*((word32*)block[j]) = *((word32*)(input+4*j));
	}
	switch (key->direction) {
	case DIR_ENCRYPT:
		rijndaelEncryptRound (block, key->keyLen, cipher->blockLen, 
				key->keySched, rounds);
	break;
		
	case DIR_DECRYPT:
		rijndaelDecryptRound (block, key->keyLen, cipher->blockLen, 
				key->keySched, rounds);
	break;
		
	default: return BAD_KEY_DIR;
	} 
	for (j = cipher->blockLen/32-1; j >= 0; j--) {
		/* parse rectangular array into output ciphertext bytes */
		*((word32*)(outBuffer+4*j)) = *((word32*)block[j]);
	}
	
	return TRUE;
}
