/*******************************************************************************
*
* FILE:         speedtst.c
*
* DESCRIPTION:  main program to measure the speed of Crypton
*
* AUTHOR:       Chae Hoon Lim and Hyo Sun Hwang
*               E-mail: chlim@future.co.kr, hyosun@future.co.kr
*               http://crypt.future.co.kr/~chlim
*               Information and Communications Research Center,
*               Future Systems, Inc.
*               372-2, Yang Jae-Dong, Seo Cho-Gu, Seoul, 137-130, KOREA
*               Tel: +82-2-578-0581 (ext. 557), Fax: +82-2-578-0584
*
* DATE:         July 10, 1998
*
* NOTE: This code must be compiled and linked with "crypton.o" and "aesapi.o"
*       to produce the exe file "speedtst.exe".
*
*******************************************************************************/

#include <time.h>
#include "aesapi.h"

#if defined(CheckInterValue)
    #error DO NOT define 'CheckInterValue' for speed test.
#endif

#define BASE_ITER       1000

/***************************** Common Macros *******************************/

#define SPEED(str, oper, wt, t) {         \
    clock_t start, finish;                \
    double diff_time;                     \
    DWORD idx;                            \
    start = clock();                      \
    for( idx=0; idx<BASE_ITER*wt; idx++)  \
        oper;                             \
    finish = clock();                     \
    diff_time = ((double)finish - (double)start)/CLOCKS_PER_SEC;         \
    printf("\n%18s - %5ldK Iterations", str, idx/1000);                  \
    printf(" => %5ld cycles", (DWORD)(1.0e6*diff_time/idx*CPU_MHZ) );    \
    if(t==1)                                                             \
       printf(" (%4.2f Mbyte/sec)", (double)16*idx/diff_time/1024/1024); \
}


#define BytePrint(FP, msg, DataSize, Data) {                  \
    DWORD idx;                                                \
    fprintf(FP, "%s\t", msg);                                 \
    for( idx=0; idx<(DWORD)DataSize; idx++) {                 \
        if( (idx!=0) && (idx%16==0) ) fprintf(FP, "\n\t\t");  \
        fprintf(FP, " %.2X", ((BYTE *)Data)[idx]);            \
    }                                                         \
    fprintf(FP, "\n");                                        \
}


/******************************** Test Speed *******************************/

int main(argc, argv)
int argc;
char *argv[];
{
    int i, KeyLen;
    BYTE UK[MAX_KEY_SIZE], PT[BLOCK_SIZE],
         CT[BLOCK_SIZE], CT2[BLOCK_SIZE], DT[BLOCK_SIZE];
    keyInstance key;
    cipherInstance cipher;
    DWORD ExpandedKey[NoExpandedKeys];
    DWORD EncRoundKey[NoRoundKeys], DecRoundKey[NoRoundKeys];

#if defined(USE_ASM)
    printf("Using Inline Assemly");
#else
    printf("Using C only");
#endif

#if defined(BIG_ENDIAN)
    printf(", BIG_ENDIAN defined");
#else
    printf(", LITTLE_ENDIAN defined");
#endif

    printf(", %d MHz clock rate assumed.", CPU_MHZ);

    for( KeyLen=128; KeyLen<=256; KeyLen+=64) {
        for( i=0; i<KeyLen/8; i++)      UK[i] = i;
        for( i=0; i<BLOCK_SIZE; i++)    PT[i] = i;
        printf("\nKey Length = %d:", KeyLen);

        /*  */
        SPEED("EMPTY", {}, 30000, 0);

        /* measure speed for AES-API calls */
        SPEED("makeKey", makeKey(&key, DIR_ENCRYPT, KeyLen, UK), 180, 0);
        SPEED("cipherInit", cipherInit(&cipher, MODE_ECB, NULL), 6000, 0);
        SPEED("blockEncrypt", blockEncrypt(&cipher, &key, PT, BLOCK_LEN, CT), 180, 1);

        SPEED("makeKey", makeKey(&key, DIR_DECRYPT, KeyLen, UK), 180, 0);
        SPEED("cipherInit", cipherInit(&cipher, MODE_ECB, NULL), 6000, 0);
        SPEED("blockDecrypt", blockDecrypt(&cipher, &key, CT, BLOCK_LEN, DT), 160, 1);

        if( memcmp(PT, DT, BLOCK_SIZE)==0 ) printf("(OK)");
        else                                printf("(FAIL)");

        /* measure speed for CRYPTON core functions */
        SPEED("CryptonGenSbox", CryptonGenSbox(), 7, 0);
        SPEED("CryptonExpandKey", CryptonExpandKey(ExpandedKey, UK, KeyLen/8), 300, 0);
        SPEED("GenEncRoundKey", CryptonGenEncRoundKey(EncRoundKey, ExpandedKey), 600, 0);
        SPEED("GenDecRoundKey", CryptonGenDecRoundKey(DecRoundKey, ExpandedKey), 600, 0);

        memcpy(CT2, PT, BLOCK_SIZE);
        SPEED("CryptonEncrypt", CryptonEncrypt(CT2, EncRoundKey), 200, 1);
        memcpy(DT, CT2, BLOCK_SIZE);
        SPEED("CryptonEncrypt", CryptonEncrypt(DT, DecRoundKey), 200, 1);

        if( memcmp(PT, DT, BLOCK_SIZE)==0 ) printf("(OK)\n");
        else                                printf("(FAIL)\n");

        BytePrint(stdout, "\tUK", KeyLen/8, UK);
        BytePrint(stdout, "\tC1", BLOCK_SIZE, CT);
        BytePrint(stdout, "\tC2", BLOCK_SIZE, CT2);
    }
    printf("\n");

	return 1;
}

/********************************** END **************************************/
