package Crypton;

import java.io.PrintWriter;
import java.security.InvalidKeyException;

/****************************************************************************
*             Java implementation of CRYPTON Version 0.5 
*****************************************************************************
*
* Crypton is a 128-bit block cipher with variable key size 
* (64+8k, k=0,1,...,24), designed by Chae Hoon Lim, Future Systems, Inc.
* Email: chlim@future.co.kr
*
* This Java code was written by Chae Hoon Lim and Myung Hee Kang, 
* Future Systems, Inc. Email: mhkang@future.co.kr
*
****************************************************************************/


public final class Crypton_Algorithm // implicit no-argument constructor
{
    /* Debugging methods and variables */
    static final String NAME = "Crypton_Algorithm";
    static final boolean IN = true, OUT = false;

    static final boolean DEBUG = Crypton_Properties.GLOBAL_DEBUG;
    static final int debuglevel = DEBUG ? Crypton_Properties.getLevel(NAME) : 0;
    static final PrintWriter err = DEBUG ? Crypton_Properties.getOutput() : null;

    static final boolean TRACE = Crypton_Properties.isTraceable(NAME);

    static void debug (String s) { err.println(">>> "+NAME+": "+s); }
    static void trace (boolean in, String s) {
        if (TRACE) err.println((in?"==> ":"<== ")+NAME+"."+s);
    }
    static void trace (String s) { if (TRACE) err.println("<=> "+NAME+"."+s); }


    /* Constants and variables */
    static final int BLOCK_SIZE = 16; // bytes in a data-block
    static final int NoExpandedKeys  = 8;
    
    /* P-boxes used to construct S-boxes, S0 and S1 */
    static final byte[] P0 = new byte[] {
                15,  9,  6,  8,  9,  9,  4, 12,
                 6,  2,  6, 10,  1,  3,  5, 15 };
    
    static final byte[] P1 = new byte[] {
            10, 15,  4,  7,  5,  2, 14,  6,
             9,  3, 12,  8, 13,  1, 11,  0 };

    static final byte[] P2 = new byte[] {
                 0,  4,  8,  4,  2, 15,  8, 13,
                 1,  1, 15,  7,  2, 11, 14, 15 };
    
    /* S-boxes created from P-boxes */
    static final byte[] S0 = new byte[256];
    static final byte[] S1 = new byte[256];

    /* SS-boxes used for optimized implementation via table lookup */
    static final  int[] SS0 = new int[256];
    static final  int[] SS1 = new int[256];
    static final  int[] SS2 = new int[256];
    static final  int[] SS3 = new int[256];
    
    private static final char[] HEX_DIGITS = {
            '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
        };


/***************************************************************************/

    /* Static code - to intialise S-boxes and SS-boxes */
    static {

        long time = System.currentTimeMillis();

        if (DEBUG && debuglevel > 6) {
            System.out.println("Algorithm Name: "+Crypton_Properties.FULL_NAME);
            System.out.println("Electronic Codebook (ECB) Mode");
            System.out.println();
        }

        int i, j, R, L;

        for (i = 0; i < 256; i++)
        {
            R =  i & 0xf;
            L = ( i >>>  4 ) ^ P0[R];
            R ^= P1[L];
            L ^= P2[R];
            R =  ( L << 4 ) ^ R;
             S0[i] = (byte) R;
            SS0[i] = ( ( R & 0x3f )       ) | ( ( R & 0xcf ) <<  8 ) |
                     ( ( R & 0xf3 ) << 16 ) | ( ( R & 0xfc ) << 24 );

            SS2[i] = ( ( R & 0xf3 )       ) | ( ( R & 0xfc ) <<  8 ) |
                     ( ( R & 0x3f ) << 16 ) | ( ( R & 0xcf ) << 24 );

            S1[R] = (byte) i; 
            SS1[R] = ( ( i & 0xfc )       ) | ( ( i & 0x3f ) <<  8 ) |
                     ( ( i & 0xcf ) << 16 ) | ( ( i & 0xf3 ) << 24 );
            SS3[R] = ( ( i & 0xcf )       ) | ( ( i & 0xf3 ) <<  8 ) |
                     ( ( i & 0xfc ) << 16 ) | ( ( i & 0x3f ) << 24 );
        }

        time = System.currentTimeMillis() - time;

        if (DEBUG && debuglevel > 8) {
            System.out.println("==========");
            System.out.println();
            System.out.println("Static Data");
            System.out.println();

            System.out.println("S0[]:");
            for(i=0;i<16;i++) {
                for(j=0;j<16;j++)
                    System.out.print("0x"+byteToString(S0[i*16+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println("S1[]:");
            for(i=0;i<16;i++) {
                for(j=0;j<16;j++)
                    System.out.print("0x"+byteToString(S1[i*16+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println("SS0[]:");
            for(i=0;i<32;i++) {
                for(j=0;j<8;j++)
                    System.out.print("0x"+intToString(SS0[i*8+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println("SS1[]:");
            for(i=0;i<32;i++) {
                for(j=0;j<8;j++)
                    System.out.print("0x"+intToString(SS1[i*8+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println("SS2[]:");
            for(i=0;i<32;i++) {
                for(j=0;j<8;j++)
                    System.out.print("0x"+intToString(SS2[i*8+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println("SS3[]:");
            for(i=0;i<32;i++) {
                for(j=0;j<8;j++)
                    System.out.print("0x"+intToString(SS3[i*8+j])+" ");
                System.out.println();
            }
            System.out.println();
    
            System.out.println();
            System.out.println("Total initialization time: "+time+" ms.");
            System.out.println();
        }

    }

/**************************************************************************/

    /* masking values for bit permutation */ 
    static final int MW0 = 0xfcf3cf3f;
    static final int MW1 = 0xf3cf3ffc;
    static final int MW2 = 0xcf3ffcf3;
    static final int MW3 = 0x3ffcf3cf;
        
    /* Key expansion : Bit permutation, constant addition,
                 substitution and byte transposition */
    static final void KeyExpand0(int[] R, int[] T, int K0, int K1, int K2, int K3)
    {   
        int A, B, C, D;
        
        A    = (T[0] & MW0) ^ (T[1] & MW1) ^ (T[2] & MW2) ^ (T[3] & MW3) ^ K0;
        B    = (T[0] & MW1) ^ (T[1] & MW2) ^ (T[2] & MW3) ^ (T[3] & MW0) ^ K1;
        C    = (T[0] & MW2) ^ (T[1] & MW3) ^ (T[2] & MW0) ^ (T[3] & MW1) ^ K2;
        D    = (T[0] & MW3) ^ (T[1] & MW0) ^ (T[2] & MW1) ^ (T[3] & MW2) ^ K3;

        R[0] = ( (S0[(A >>> 24) & 0xff] & 0xff) << 24 ) ^ 
               ( (S1[(B >>> 24) & 0xff] & 0xff) << 16 ) ^
               ( (S0[(C >>> 24) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S1[(D >>> 24) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[1] = ( (S1[(A >>> 16) & 0xff] & 0xff) << 24 ) ^ 
               ( (S0[(B >>> 16) & 0xff] & 0xff) << 16 ) ^
               ( (S1[(C >>> 16) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S0[(D >>> 16) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[2] = ( (S0[(A >>>  8) & 0xff] & 0xff) << 24 ) ^ 
               ( (S1[(B >>>  8) & 0xff] & 0xff) << 16 ) ^
               ( (S0[(C >>>  8) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S1[(D >>>  8) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[3] = ( (S1[(A       ) & 0xff] & 0xff) << 24 ) ^ 
               ( (S0[(B       ) & 0xff] & 0xff) << 16 ) ^
               ( (S1[(C       ) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S0[(D       ) & 0xff] & 0xff)          ) ;
    }

    static final void KeyExpand1(int[] R, int[] T, int K0, int K1, int K2, int K3)
    {
        int A, B, C, D;
        
        A    = (T[0] & MW1) ^ (T[1] & MW2) ^ (T[2] & MW3) ^ (T[3] & MW0) ^ K0;
        B    = (T[0] & MW2) ^ (T[1] & MW3) ^ (T[2] & MW0) ^ (T[3] & MW1) ^ K1;
        C    = (T[0] & MW3) ^ (T[1] & MW0) ^ (T[2] & MW1) ^ (T[3] & MW2) ^ K2;
        D    = (T[0] & MW0) ^ (T[1] & MW1) ^ (T[2] & MW2) ^ (T[3] & MW3) ^ K3;

        R[0] = ( (S1[(A >>> 24) & 0xff] & 0xff) << 24 ) ^ 
               ( (S0[(B >>> 24) & 0xff] & 0xff) << 16 ) ^
               ( (S1[(C >>> 24) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S0[(D >>> 24) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[1] = ( (S0[(A >>> 16) & 0xff] & 0xff) << 24 ) ^ 
               ( (S1[(B >>> 16) & 0xff] & 0xff) << 16 ) ^
               ( (S0[(C >>> 16) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S1[(D >>> 16) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[2] = ( (S1[(A >>>  8) & 0xff] & 0xff) << 24 ) ^ 
               ( (S0[(B >>>  8) & 0xff] & 0xff) << 16 ) ^
               ( (S1[(C >>>  8) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S0[(D >>>  8) & 0xff] & 0xff)          ) ;
                                                                                                  
        R[3] = ( (S0[(A       ) & 0xff] & 0xff) << 24 ) ^ 
               ( (S1[(B       ) & 0xff] & 0xff) << 16 ) ^
               ( (S0[(C       ) & 0xff] & 0xff) <<  8 ) ^ 
               ( (S1[(D       ) & 0xff] & 0xff)          ) ;
    }

/************************************************************************/


    /* constants for key expansion */
    static final int   C0 = 0x85ae67bb;
    static final int   C1 = 0x72f36e3c;
    static final int   C2 = 0x3af54fa5;
    static final int   C3 = 0x7f520e51;
    static final int   C4 = 0x8c68059b;
    static final int   C5 = 0xabd9831f;
    static final int   C6 = 0x19cde05b;
    static final int   C7 = 0x5d9dbbcb;

    /* Generate 8 Expanded keys from a user-supplied key */
    static final void CryptonExpandKey(int[] pdwExpKey, byte[] pbUserKey, int dwUserKeyLen)
    {
        int i;
        int SUM;

        int[]  UK  = new  int[NoExpandedKeys];

        int[]  EK0 = new int[NoExpandedKeys / 2];
        int[]  EK1 = new int[NoExpandedKeys / 2];

        int[]   T  = new int[4];

        for (i = 0; i < dwUserKeyLen / 4; i++)
           UK[i] = ( (pbUserKey[i*4 + 0] & 0xff) << 24 ) ^
                   ( (pbUserKey[i*4 + 1] & 0xff) << 16 ) ^
                   ( (pbUserKey[i*4 + 2] & 0xff) <<  8 ) ^
                   ( (pbUserKey[i*4 + 3] & 0xff)       ) ;

        for(; i < NoExpandedKeys; i++)
            UK[i] = 0;
    
        T[0] = UK[0];
        T[1] = UK[2];
        T[2] = UK[4];
        T[3] = UK[6];
        KeyExpand0(EK0, T, C0, C1, C2, C3);
        SUM = EK0[0] ^ EK0[1] ^ EK0[2] ^ EK0[3];

        T[0] = UK[1];
        T[1] = UK[3];
        T[2] = UK[5];
        T[3] = UK[7];

        KeyExpand1(EK1, T, C4, C5, C6, C7);

        pdwExpKey[4] = EK1[0] ^ SUM;
        pdwExpKey[5] = EK1[1] ^ SUM;
        pdwExpKey[6] = EK1[2] ^ SUM;
        pdwExpKey[7] = EK1[3] ^ SUM;

        SUM = pdwExpKey[4] ^ pdwExpKey[5] ^ pdwExpKey[6] ^ pdwExpKey[7];

        pdwExpKey[0] = EK0[0] ^ SUM;
        pdwExpKey[1] = EK0[1] ^ SUM;
        pdwExpKey[2] = EK0[2] ^ SUM;
        pdwExpKey[3] = EK0[3] ^ SUM;
    }

/***********************************************************************/
    
   /* masking value for word permutation */
    static final int  M20 = 0xcffccffc;
    static final int  M31 = 0x3ff33ff3;
    static final int  M02 = 0xfccffccf;
    static final int  M13 = 0xf33ff33f;

    /* rotate-right of a 32-bit number */
    static final int ROTR(int x, int n)
    {
        return ( (x >>> n) | (x << 32 - n) ) ;
    }

    /* word permutations */
    static final int WordPerm0(int A0)
    {
        return (      (A0 & M02)    ) ^ ROTR( (A0 & M31), 24) ^ 
               ROTR( (A0 & M20), 16) ^ ROTR( (A0 & M13),  8);
    }

    static final int WordPerm1(int A1)
    {
        return (      (A1 & M13)    ) ^ ROTR( (A1 & M02), 24) ^ 
               ROTR( (A1 & M31), 16) ^ ROTR( (A1 & M20), 8);
    }

    static final int WordPerm2(int A2)
    {
        return (      (A2 & M20)    ) ^ ROTR( (A2 & M13), 24) ^ 
               ROTR( (A2 & M02), 16) ^ ROTR( (A2 & M31),  8);
    }

    static final int WordPerm3(int A3)
    {
        return (      (A3 & M31)    ) ^ ROTR( (A3 & M20), 24) ^ 
               ROTR( (A3 & M13), 16) ^ ROTR( (A3 & M02),  8);
    }

   /* round constants */
    static final int RC0 = 0x01010101;
    static final int RC1 = 0x02020202;
    static final int RC2 = 0x04040404;
    static final int RC3 = 0x08080808;
    static final int RC4 = 0x10101010;
    static final int RC5 = 0x20202020;

    /* generate encryption round keys from Expanded keys */
    static final void CryptonGenEncRoundKey(int[] pdwEncKey, int[] pdwExpKey)
    {
        pdwEncKey[ 0] = pdwExpKey[0]              ;
        pdwEncKey[ 8] = ROTR(pdwExpKey[ 0],  8) ;
        pdwEncKey[16] = RC1 ^ pdwEncKey [ 8]      ;
        pdwEncKey[24] = ROTR(pdwEncKey[16], 16) ;
        pdwEncKey[32] = RC3 ^ pdwEncKey [24]      ;
        pdwEncKey[40] = ROTR(pdwEncKey[32], 24) ;
        pdwEncKey[48] = RC5 ^ pdwEncKey [40]      ;
        pdwEncKey[48] = WordPerm1(pdwEncKey[48])  ;

        pdwEncKey[ 1] = pdwExpKey[1]              ;
        pdwEncKey[ 9] = RC0 ^ pdwEncKey [ 1]      ;
        pdwEncKey[17] = ROTR(pdwEncKey[ 9], 24) ;
        pdwEncKey[25] = RC2 ^ pdwEncKey [17]      ;
        pdwEncKey[33] = ROTR(pdwEncKey[25],  8) ;
        pdwEncKey[41] = RC4 ^ pdwEncKey [33]      ;
        pdwEncKey[49] = ROTR(pdwEncKey[41], 16) ;
        pdwEncKey[49] = WordPerm2(pdwEncKey[49])  ;

        pdwEncKey[ 2] = pdwExpKey[2]              ;
        pdwEncKey[10] = ROTR(pdwEncKey[ 2], 16) ;
        pdwEncKey[18] = RC1 ^ pdwEncKey [10]      ;
        pdwEncKey[26] = ROTR(pdwEncKey[18], 24) ;
        pdwEncKey[34] = RC3 ^ pdwEncKey [26]      ;
        pdwEncKey[42] = ROTR(pdwEncKey[34],  8) ;
        pdwEncKey[50] = RC5 ^ pdwEncKey [42]      ;
        pdwEncKey[50] = WordPerm3(pdwEncKey[50])  ;

        pdwEncKey[ 3] = pdwExpKey[3]              ;
        pdwEncKey[11] = RC0 ^ pdwEncKey [ 3]      ;
        pdwEncKey[19] = ROTR(pdwEncKey[11],  8) ;
        pdwEncKey[27] = RC2 ^ pdwEncKey [19]      ;
        pdwEncKey[35] = ROTR(pdwEncKey[27], 16) ;
        pdwEncKey[43] = RC4 ^ pdwEncKey [35]      ;
        pdwEncKey[51] = ROTR(pdwEncKey[43], 24) ;
        pdwEncKey[51] = WordPerm0(pdwEncKey[51])  ;

        pdwEncKey[ 4] = pdwExpKey[4]              ;
        pdwEncKey[12] = RC0 ^ pdwEncKey [ 4]      ;
        pdwEncKey[20] = ROTR(pdwEncKey[12],  8) ;
        pdwEncKey[28] = RC2 ^ pdwEncKey [20]      ;
        pdwEncKey[36] = ROTR(pdwEncKey[28], 16) ;
        pdwEncKey[44] = RC4 ^ pdwEncKey [36]      ;
    
        pdwEncKey[ 5] = pdwExpKey[5]              ;
        pdwEncKey[13] = ROTR(pdwEncKey[ 5], 16) ;
        pdwEncKey[21] = RC1 ^ pdwEncKey [13]      ;
        pdwEncKey[29] = ROTR(pdwEncKey[21], 24) ;
        pdwEncKey[37] = RC3 ^ pdwEncKey [29]      ;
        pdwEncKey[45] = ROTR(pdwEncKey[37],  8) ;
    
        pdwEncKey[ 6] = pdwExpKey[6]              ;
        pdwEncKey[14] = RC0 ^ pdwEncKey [ 6]      ;
        pdwEncKey[22] = ROTR(pdwEncKey[14], 16) ;
        pdwEncKey[30] = RC2 ^ pdwEncKey [22]      ;
        pdwEncKey[38] = ROTR(pdwEncKey[30], 24) ;
        pdwEncKey[46] = RC4 ^ pdwEncKey [38]      ;
    
        pdwEncKey[ 7] = pdwExpKey[7]              ;
        pdwEncKey[15] = ROTR(pdwEncKey[ 7], 24) ;
        pdwEncKey[23] = RC1 ^ pdwEncKey [15]      ;
        pdwEncKey[31] = ROTR(pdwEncKey[23],  8) ;
        pdwEncKey[39] = RC3 ^ pdwEncKey [31]      ;
        pdwEncKey[47] = ROTR(pdwEncKey[39], 16) ;
    }

    /* generate decryption round keys from Expanded keys */
    static final void CryptonGenDecRoundKey(int[] pdwDecKey, int[] pdwExpKey)
    {
        pdwDecKey[48] = pdwExpKey[0]              ;
        pdwDecKey[40] = WordPerm1(pdwExpKey[ 0] ) ;
        pdwDecKey[40] = ROTR(pdwDecKey[40], 24) ;
        pdwDecKey[32] = RC1 ^ pdwDecKey [40]      ;
        pdwDecKey[24] = ROTR(pdwDecKey[32], 16) ;
        pdwDecKey[16] = RC3^ pdwDecKey  [24]      ;
        pdwDecKey[ 8] = ROTR(pdwDecKey[16],  8) ;
        pdwDecKey[ 0] = RC5 ^ pdwDecKey[ 8]       ;
 
        pdwDecKey[49] = pdwExpKey[1]              ;
        pdwDecKey[41] = WordPerm2(pdwExpKey[ 1] ) ;
        pdwDecKey[41] = RC0 ^ pdwDecKey [41]      ;
        pdwDecKey[33] = ROTR(pdwDecKey[41],  8) ;
        pdwDecKey[25] = RC2 ^ pdwDecKey [33]      ;
        pdwDecKey[17] = ROTR(pdwDecKey[25], 24) ;
        pdwDecKey[ 9] = RC4 ^ pdwDecKey [17]      ;
        pdwDecKey[ 1] = ROTR(pdwDecKey[ 9], 16) ;
 
        pdwDecKey[50] = pdwExpKey[2]              ;
        pdwDecKey[42] = WordPerm3(pdwExpKey[ 2] ) ;
        pdwDecKey[42] = ROTR(pdwDecKey[42], 16) ;
        pdwDecKey[34] = RC1 ^ pdwDecKey [42]      ;
        pdwDecKey[26] = ROTR(pdwDecKey[34],  8) ;
        pdwDecKey[18] = RC3 ^ pdwDecKey [26]      ;
        pdwDecKey[10] = ROTR(pdwDecKey[18], 24) ;
        pdwDecKey[ 2] = RC5 ^ pdwDecKey [10]      ;
 
        pdwDecKey[51] = pdwExpKey[3]              ;
        pdwDecKey[43] = WordPerm0(pdwExpKey[ 3] ) ;
        pdwDecKey[43] = RC0 ^ pdwDecKey [43]      ;
        pdwDecKey[35] = ROTR(pdwDecKey[43], 24) ;
        pdwDecKey[27] = RC2 ^ pdwDecKey [35]      ;
        pdwDecKey[19] = ROTR(pdwDecKey[27], 16) ;
        pdwDecKey[11] = RC4 ^ pdwDecKey [19]      ;
        pdwDecKey[ 3] = ROTR(pdwDecKey[11],  8) ;
 
        pdwDecKey[44] = WordPerm0(pdwExpKey[ 4] ) ;
        pdwDecKey[36] = RC0 ^ pdwDecKey [44]      ;
        pdwDecKey[28] = ROTR(pdwDecKey[36], 24) ;
        pdwDecKey[20] = RC2 ^ pdwDecKey [28]      ;
        pdwDecKey[12] = ROTR(pdwDecKey[20], 16) ;
        pdwDecKey[ 4] = RC4 ^ pdwDecKey [12]      ;
     
        pdwDecKey[45] = WordPerm1(pdwExpKey[ 5] ) ;
        pdwDecKey[37] = ROTR(pdwDecKey[45], 16) ;
        pdwDecKey[29] = RC1 ^ pdwDecKey [37]      ;
        pdwDecKey[21] = ROTR(pdwDecKey[29],  8) ;
        pdwDecKey[13] = RC3 ^ pdwDecKey [21]      ;
        pdwDecKey[ 5] = ROTR(pdwDecKey[13], 24) ;
     
        pdwDecKey[46] = WordPerm2(pdwExpKey[ 6] ) ;
        pdwDecKey[38] = RC0 ^ pdwDecKey [46]      ;
        pdwDecKey[30] = ROTR(pdwDecKey[38], 16) ;
        pdwDecKey[22] = RC2 ^ pdwDecKey [30]      ;
        pdwDecKey[14] = ROTR(pdwDecKey[22],  8) ;
        pdwDecKey[ 6] = RC4 ^ pdwDecKey [14]      ;
     
        pdwDecKey[47] = WordPerm3(pdwExpKey[ 7] ) ;
        pdwDecKey[39] = ROTR(pdwDecKey[47],  8) ;
        pdwDecKey[31] = RC1 ^ pdwDecKey [39]      ;
        pdwDecKey[23] = ROTR(pdwDecKey[31], 24) ;
        pdwDecKey[15] = RC3 ^ pdwDecKey [23]      ;
        pdwDecKey[ 7] = ROTR(pdwDecKey[15], 16) ;
    }

/***********************************************************************/

    /* one-block encryption/decryption */
    static final void Crypton(byte[] in, int[] Kr, byte[] result)
    {
        int A, B, C, D, E, F, G, H;

        A = ( (in[ 0] & 0xff) << 24 | (in[ 1] & 0xff) << 16 |
              (in[ 2] & 0xff) <<  8 | (in[ 3] & 0xff)        ) ^ Kr[0];
                                                                      
        B = ( (in[ 4] & 0xff) << 24 | (in[ 5] & 0xff) << 16 |
              (in[ 6] & 0xff) <<  8 | (in[ 7] & 0xff)        ) ^ Kr[1];
                                                                      
        C = ( (in[ 8] & 0xff) << 24 | (in[ 9] & 0xff) << 16 |
              (in[10] & 0xff) <<  8 | (in[11] & 0xff)        ) ^ Kr[2];
                                                                      
        D = ( (in[12] & 0xff) << 24 | (in[13] & 0xff) << 16 |
              (in[14] & 0xff) <<  8 | (in[15] & 0xff)        ) ^ Kr[3];

        if (DEBUG && debuglevel > 6)
            System.out.println("CT{PT}.R"+" = "+intToString(A)+intToString(B)+
                                    intToString(C)+intToString(D) );

        for(int r = 4; r < 44; r += 8) {
            E = ( SS0[(A >>> 24) & 0xff] ^ SS1[(B >>> 24) & 0xff] ^
                  SS2[(C >>> 24) & 0xff] ^ SS3[(D >>> 24) & 0xff] ) ^ Kr[r];

            F = ( SS1[(A >>> 16) & 0xff] ^ SS2[(B >>> 16) & 0xff] ^
                  SS3[(C >>> 16) & 0xff] ^ SS0[(D >>> 16) & 0xff] ) ^ Kr[r+1];

            G = ( SS2[(A >>>  8) & 0xff] ^ SS3[(B >>>  8) & 0xff] ^
                  SS0[(C >>>  8) & 0xff] ^ SS1[(D >>>  8) & 0xff] ) ^ Kr[r+2];

            H = ( SS3[(A       ) & 0xff] ^ SS0[(B       ) & 0xff] ^
                  SS1[(C       ) & 0xff] ^ SS2[(D       ) & 0xff] ) ^ Kr[r+3];

            if (DEBUG && debuglevel > 6)
                System.out.println("CT{PT}.R"+" = "+intToString(E)+
                    intToString(F)+intToString(G)+intToString(H) );

            A = ( SS1[(E >>> 24) & 0xff] ^ SS2[(F >>> 24) & 0xff ] ^
                  SS3[(G >>> 24) & 0xff] ^ SS0[(H >>> 24) & 0xff] ) ^ Kr[r+4];

            B = ( SS2[(E >>> 16) & 0xff] ^ SS3[(F >>> 16) & 0xff ] ^
                  SS0[(G >>> 16) & 0xff] ^ SS1[(H >>> 16) & 0xff] ) ^ Kr[r+5];

            C = ( SS3[(E >>>  8) & 0xff] ^ SS0[(F >>>  8) & 0xff ] ^
                  SS1[(G >>>  8) & 0xff] ^ SS2[(H >>>  8) & 0xff ] ) ^ Kr[r+6];

            D = ( SS0[(E       ) & 0xff] ^ SS1[(F       ) & 0xff] ^
                  SS2[(G       ) & 0xff] ^ SS3[(H       ) & 0xff ] ) ^ Kr[r+7];                                              

            if (DEBUG && debuglevel > 6)
                System.out.println("CT{PT}.R"+" = "+intToString(A)+
                    intToString(B)+intToString(C)+intToString(D) );
        }

        E = ( SS0[(A >>> 24) & 0xff] ^ SS1[(B >>> 24) & 0xff] ^
              SS2[(C >>> 24) & 0xff] ^ SS3[(D >>> 24) & 0xff] ) ^ Kr[44];
               
        F = ( SS1[(A >>> 16) & 0xff] ^ SS2[(B >>> 16) & 0xff] ^
              SS3[(C >>> 16) & 0xff] ^ SS0[(D >>> 16) & 0xff] ) ^ Kr[45];
                
        G = ( SS2[(A >>>  8) & 0xff] ^ SS3[(B >>>  8) & 0xff] ^
              SS0[(C >>>  8) & 0xff] ^ SS1[(D >>>  8) & 0xff] ) ^ Kr[46];
                
        H = ( SS3[(A       ) & 0xff] ^ SS0[(B       ) & 0xff] ^
              SS1[(C       ) & 0xff] ^ SS2[(D       ) & 0xff] ) ^ Kr[47];
              
        if (DEBUG && debuglevel > 6)
            System.out.println("CT{PT}.R"+" = "+intToString(E)+intToString(F)+
                                    intToString(G)+intToString(H) );

        result[ 0] = (byte)( S1[(E >>> 24) & 0xff] ^ (Kr[48] >>> 24) );
        result[ 1] = (byte)( S0[(F >>> 24) & 0xff] ^ (Kr[48] >>> 16) );
        result[ 2] = (byte)( S1[(G >>> 24) & 0xff] ^ (Kr[48] >>>  8) );
        result[ 3] = (byte)( S0[(H >>> 24) & 0xff] ^ (Kr[48]       ) );
        result[ 4] = (byte)( S0[(E >>> 16) & 0xff] ^ (Kr[49] >>> 24) );
        result[ 5] = (byte)( S1[(F >>> 16) & 0xff] ^ (Kr[49] >>> 16) );
        result[ 6] = (byte)( S0[(G >>> 16) & 0xff] ^ (Kr[49] >>>  8) );
        result[ 7] = (byte)( S1[(H >>> 16) & 0xff] ^ (Kr[49]       ) );
                                                             
        result[ 8] = (byte)( S1[(E >>>  8) & 0xff] ^ (Kr[50] >>> 24) );
        result[ 9] = (byte)( S0[(F >>>  8) & 0xff] ^ (Kr[50] >>> 16) );
        result[10] = (byte)( S1[(G >>>  8) & 0xff] ^ (Kr[50] >>>  8) );
        result[11] = (byte)( S0[(H >>>  8) & 0xff] ^ (Kr[50]       ) );
        result[12] = (byte)( S0[(E         ) & 0xff] ^ (Kr[51] >>> 24) );
        result[13] = (byte)( S1[(F         ) & 0xff] ^ (Kr[51] >>> 16) );
        result[14] = (byte)( S0[(G         ) & 0xff] ^ (Kr[51] >>>  8) );
        result[15] = (byte)( S1[(H         ) & 0xff] ^ (Kr[51]       ) ); 

        if (DEBUG && debuglevel > 6)
            System.out.println("CT{PT}.R"+" = "+byteToString(result[ 0]) + 
                byteToString(result[ 1])+ byteToString(result[ 2]) + 
                byteToString(result[ 3])+ byteToString(result[ 4]) + 
                byteToString(result[ 5])+ byteToString(result[ 6]) + 
                byteToString(result[ 7])+ byteToString(result[ 8]) + 
                byteToString(result[ 9])+ byteToString(result[10]) + 
                byteToString(result[11])+ byteToString(result[12]) + 
                byteToString(result[13])+ byteToString(result[14]) + 
                byteToString(result[15]) );
    }

/************************************************************************/

    /************ Basic AES API calls *************/

    /* Expand a user-supplied key material into a session key.
     * @param key  The (64+8k)-bit user-key to use.
     * @return  This cipher's round keys.
     * @exception  InvalidKeyException  If the key is invalid. */
    public static synchronized Object makeKey (byte[] k)
    throws InvalidKeyException {
        if (DEBUG) trace(IN, "makeKey("+k+")");
        if (k == null)
            throw new InvalidKeyException("Empty key");
        if ((k.length <  8) || (k.length > 32))
            throw new InvalidKeyException("Incorrect key length");

        int[]  ExpKey = new int[8];

        int[]  Ke = new int[52];
        int[]  Kd = new int[52];

        /* generate expanded keys from a user key */
        CryptonExpandKey(ExpKey, k, k.length);
        /* generate encryption round keys from expanded keys */
        CryptonGenEncRoundKey(Ke, ExpKey);

        if (DEBUG && debuglevel > 7) {
            System.out.println("Raw    = "+toString(k));
            System.out.println("ExpKey"+" = "+
                    intToString(ExpKey[ 0])+" "+
                    intToString(ExpKey[ 1])+" "+
                    intToString(ExpKey[ 2])+" "+
                    intToString(ExpKey[ 3]));
            System.out.println("      "+" = "+
                    intToString(ExpKey[ 4])+" "+
                    intToString(ExpKey[ 5])+" "+
                    intToString(ExpKey[ 6])+" "+
                    intToString(ExpKey[ 7]));
            System.out.println();
            System.out.println("EncRK "+" = "+
                    intToString(Ke[0])+" "+
                    intToString(Ke[1])+" "+
                    intToString(Ke[2])+" "+
                    intToString(Ke[3]));
            for (int i = 4; i < 52; i += 4) 
            System.out.println("      "+" = "+
                    intToString(Ke[i])+" "+
                    intToString(Ke[i+1])+" "+
                    intToString(Ke[i+2])+" "+
                    intToString(Ke[i+3]));
            System.out.println();
        }
        /* generate decryption round keys from expanded keys */
        CryptonGenDecRoundKey(Kd, ExpKey);

        if (DEBUG && debuglevel > 7) {
            System.out.println("ExpKey"+" = "+
                    intToString(ExpKey[ 0])+" "+
                    intToString(ExpKey[ 1])+" "+
                    intToString(ExpKey[ 2])+" "+
                    intToString(ExpKey[ 3]));
            System.out.println("      "+" = "+
                    intToString(ExpKey[ 4])+" "+
                    intToString(ExpKey[ 5])+" "+
                    intToString(ExpKey[ 6])+" "+
                    intToString(ExpKey[ 7]));
            System.out.println();
            System.out.println("DecRK "+" = "+
                    intToString(Kd[0])+" "+
                    intToString(Kd[1])+" "+
                    intToString(Kd[2])+" "+
                    intToString(Kd[3]));
            for (int i = 4; i < 52; i += 4) 
            System.out.println("      "+" = "+
                    intToString(Kd[i])+" "+
                    intToString(Kd[i+1])+" "+
                    intToString(Kd[i+2])+" "+
                    intToString(Kd[i+3]));
            System.out.println();
        }

    /* assemble the encryption (Ke) and decryption (Kd) round keys into
     one sessionKey object */
        Object[] sessionKey = new Object[] {Ke, Kd};

    if (DEBUG) trace(OUT, "makeKey()");
        return sessionKey;
    }

/************************************************************************/

    /**
     * Encrypt exactly one block of plaintext.
     * @param in          The plaintext.
     * @param inOffset    Index of in from which to start considering data.
     * @param sessionKey  The session key to use for encryption.
     * @return The ciphertext generated from a plaintext using the session key.
     */
    public static byte[]
    blockEncrypt (byte[] in, int inOffset, Object sessionKey) {
        if (DEBUG) trace(IN, "blockEncrypt("+in+", "+inOffset+", "+sessionKey+")");
        if (DEBUG && debuglevel > 6) {
            System.out.println("PT       = "+toString(in, inOffset, BLOCK_SIZE));
            System.out.println();
        }

        /* extract encryption round keys */
        int[] Ke = (int[]) ((Object[]) sessionKey)[0];     
  
        byte[] result = new byte[BLOCK_SIZE];

        Crypton(in, Ke, result);

        if (DEBUG && debuglevel > 6) {
            System.out.println();
            System.out.println("CT       = "+toString(result));
            System.out.println();
        }

        if (DEBUG) trace(OUT, "blockEncrypt()");
            return result;
        }

/***************************************************************************/
    /**
     * Decrypt exactly one block of ciphertext.
     * @param in          The ciphertext.
     * @param inOffset    Index of in from which to start considering data.
     * @param sessionKey  The session key to use for decryption.
     * @return The plaintext generated from a ciphertext using the session key.
     */
    public static byte[]
    blockDecrypt (byte[] in, int inOffset, Object sessionKey) {
        if (DEBUG) trace(IN, "blockDecrypt("+in+", "+inOffset+", "+sessionKey+")");
        if (DEBUG && debuglevel > 6) {
            System.out.println("CT       = "+toString(in, inOffset, BLOCK_SIZE));
            System.out.println();
        }

        /* extract decryption round keys */
        int[] Kd = (int[]) ((Object[]) sessionKey)[1];     
  
        byte[] result = new byte[BLOCK_SIZE];

        Crypton(in, Kd, result);

        if (DEBUG && debuglevel > 6) {
            System.out.println();
            System.out.println("DT       = "+toString(result));
            System.out.println();
        }

        if (DEBUG) trace(OUT, "blockDecrypt()");
            return result;
    }

/*************************************************************************/

    /** A basic symmetric encryption/decryption test. */ 
    public static boolean self_test() { return self_test(BLOCK_SIZE); }

    /** @return The length in bytes of the Algorithm input block. */
    public static int blockSize() { return BLOCK_SIZE; }

    /** A basic symmetric encryption/decryption test for a given key size. */
    private static boolean self_test (int keysize) {
        if (DEBUG) trace(IN, "self_test("+keysize+")");
            boolean ok = false;
            try {
                byte[] kb = new byte[keysize];
                byte[] pt = new byte[BLOCK_SIZE];
                int i;

                for (i = 0; i < keysize; i++)
                    kb[i] = (byte) i;
                for (i = 0; i < BLOCK_SIZE; i++)
                    pt[i] = (byte) i;

                if (DEBUG && debuglevel > 6) {
                    System.out.println("==========");
                    System.out.println();
                    System.out.println("KEYSIZE = "+(8*keysize));
                    System.out.println("KEY = "+toString(kb));
                    System.out.println();
                }
            Object key = makeKey(kb);

            if (DEBUG && debuglevel > 6) {
                System.out.println("Intermediate Ciphertext Values (Encryption)");
                System.out.println();
            }

            byte[] ct =  blockEncrypt(pt, 0, key);

            if (DEBUG && debuglevel > 6) {
                System.out.println("Intermediate Plaintext Values (Decryption)");
                System.out.println();
            }

            byte[] cpt = blockDecrypt(ct, 0, key);
            ok = areEqual(pt, cpt);
            if (!ok)
                throw new RuntimeException("Symmetric operation failed");
        } catch (Exception x) {
            if (DEBUG && debuglevel > 0) {
                debug("Exception encountered during self-test: " + x.getMessage());
                x.printStackTrace();
            }
        }
        if (DEBUG && debuglevel > 0) debug("Self-test OK? " + ok);
        if (DEBUG) trace(OUT, "self_test()");
            return ok;
    }

/**************************************************************************/

    /* utility static methods (from cryptix.util.core ArrayUtil 
        and Hex classes) */
    
    /** @return True iff the arrays have identical contents. */
    private static boolean areEqual (byte[] a, byte[] b) {
        int aLength = a.length;
        if (aLength != b.length)
            return false;
        for (int i = 0; i < aLength; i++)
            if (a[i] != b[i])
                return false;
        return true;
    }

    /**
     * Returns a string of 2 hexadecimal digits (most significant
     * digit first) corresponding to the lowest 8 bits of <i>n</i>.
     */
    private static String byteToString (int n) {
        char[] buf = {
            HEX_DIGITS[(n >>> 4) & 0x0F],
            HEX_DIGITS[ n        & 0x0F]
        };
        return new String(buf);
    }

    /**
     * Returns a string of 8 hexadecimal digits (most significant
     * digit first) corresponding to the integer <i>n</i>, which is
     * treated as unsigned.
     */
    private static String intToString (int n) {
        char[] buf = new char[8];

        for (int i = 7; i >=0; i--) {
            buf[i] = HEX_DIGITS[n & 0x0F];
            n >>>= 4;
        }
        return new String(buf);
     }

    /**
     * Returns a string of hexadecimal digits from a byte array. Each
     * byte is converted to 2 hex symbols.
     */
    private static String toString (byte[] ba) {
        return toString(ba, 0, ba.length);
    }
    private static String toString (byte[] ba, int offset, int length) {
        char[] buf = new char[length * 2];
        for (int i = offset, j = 0, k; i < offset+length; ) {
            k = ba[i++];
            buf[j++] = HEX_DIGITS[(k >>> 4) & 0x0F];
            buf[j++] = HEX_DIGITS[ k        & 0x0F];
        }
        return new String(buf);
    }

    /* main(): use to generate the Intermediate Values KAT */

    public static void main (String[] args) {
        self_test(16);
        self_test(24);
        self_test(32);
    }
}

/******************************* END ***********************************/
